package telegram

import (
	"crypto/sha1"
	"crypto/sha256"
	"fmt"
	"github.com/pkg/errors"
	"io"
	"math"
	"os"
	"path/filepath"
	"regexp"
	"strconv"
	"strings"
	"time"

	"dev.narayana.im/narayana/telegabber/xmpp/gateway"

	log "github.com/sirupsen/logrus"
	"github.com/soheilhy/args"
	"github.com/zelenin/go-tdlib/client"
)

var errOffline = errors.New("TDlib instance is offline")

var spaceRegex = regexp.MustCompile(`\s+`)
var replyRegex = regexp.MustCompile("> ?([0-9]{10,})")

const newlineChar string = "\n"

// GetContactByUsername resolves username to user id retrieves user and chat information
func (c *Client) GetContactByUsername(username string) (*client.Chat, *client.User, error) {
	if !c.online {
		return nil, nil, errOffline
	}

	chat, err := c.client.SearchPublicChat(&client.SearchPublicChatRequest{
		Username: username,
	})

	if err != nil {
		return nil, nil, err
	}

	return c.GetContactByID(chat.Id, chat)
}

// GetContactByID gets user and chat information from cache (or tries to retrieve it, if missing)
func (c *Client) GetContactByID(id int64, chat *client.Chat) (*client.Chat, *client.User, error) {
	if !c.online {
		return nil, nil, errOffline
	}

	var user *client.User
	var cacheChat *client.Chat
	var ok bool
	var err error

	if id <= math.MaxInt32 && id >= math.MinInt32 {
		userID := int32(id)
		user, ok = c.cache.users[userID]
		if !ok && userID > 0 {
			user, err = c.client.GetUser(&client.GetUserRequest{
				UserId: userID,
			})
			if err != nil {
				return nil, nil, err
			}

			c.cache.users[userID] = user
		}
	}

	cacheChat, ok = c.cache.chats[id]
	if !ok {
		if chat == nil {
			cacheChat, err = c.client.GetChat(&client.GetChatRequest{
				ChatId: id,
			})
			if err != nil {
				return nil, nil, err
			}

			c.cache.chats[id] = cacheChat
		} else {
			c.cache.chats[id] = chat
		}
	}
	if chat == nil {
		chat = cacheChat
	}

	return chat, user, nil
}

func userStatusToText(status client.UserStatus) (string, string) {
	var show, textStatus string

	switch status.UserStatusType() {
	case client.TypeUserStatusOnline:
		textStatus = "Online"
	case client.TypeUserStatusRecently:
		show, textStatus = "dnd", "Last seen recently"
	case client.TypeUserStatusLastWeek:
		show, textStatus = "unavailable", "Last seen last week"
	case client.TypeUserStatusLastMonth:
		show, textStatus = "unavailable", "Last seen last month"
	case client.TypeUserStatusEmpty:
		show, textStatus = "unavailable", "Last seen a long time ago"
	case client.TypeUserStatusOffline:
		offlineStatus, _ := status.(*client.UserStatusOffline)
		// this will stop working in 2038 O\
		elapsed := time.Now().Unix() - int64(offlineStatus.WasOnline)
		if elapsed < 3600 {
			show = "away"
		} else {
			show = "xa"
		}
		// TODO: timezone
		textStatus = time.Unix(int64(offlineStatus.WasOnline), 0).Format("Last seen at 15:04 02/01/2006")
	}

	return show, textStatus
}

func (c *Client) processStatusUpdate(chatID int64, status string, show string, args ...args.V) error {
	if !c.online {
		return nil
	}

	log.WithFields(log.Fields{
		"chat_id": chatID,
	}).Info("Status update for")

	chat, user, err := c.GetContactByID(chatID, nil)
	if err != nil {
		return err
	}

	var photo string
	if chat != nil && chat.Photo != nil {
		path := chat.Photo.Small.Local.Path
		file, err := os.Open(path)
		if err == nil {
			defer file.Close()

			hash := sha1.New()
			_, err = io.Copy(hash, file)
			if err == nil {
				photo = fmt.Sprintf("%x", hash.Sum(nil))
			} else {
				log.Errorf("Error calculating hash: %v", path)
			}
		} else if path != "" {
			log.Errorf("Photo does not exist: %v", path)
		}
	}

	if status == "" {
		if user != nil {
			show, status = userStatusToText(user.Status)
		} else {
			show, status = "chat", chat.Title
		}
	}

	gateway.SendPresence(
		c.xmpp,
		c.jid,
		gateway.SPFrom(strconv.FormatInt(chatID, 10)),
		gateway.SPShow(show),
		gateway.SPStatus(status),
		gateway.SPPhoto(photo),
		gateway.SPImmed(gateway.SPImmed.Get(args)),
	)

	return nil
}

func (c *Client) formatContact(chatID int64) string {
	if chatID == 0 {
		return ""
	}

	chat, user, err := c.GetContactByID(chatID, nil)
	if err != nil {
		return "unknown contact: " + err.Error()
	}

	var str string
	if chat != nil {
		str = fmt.Sprintf("%s (%v)", chat.Title, chat.Id)
	} else if user != nil {
		username := user.Username
		if username == "" {
			username = strconv.FormatInt(int64(user.Id), 10)
		}

		str = fmt.Sprintf("%s %s (%v)", user.FirstName, user.LastName, username)
	} else {
		str = strconv.FormatInt(chatID, 10)
	}

	str = spaceRegex.ReplaceAllString(str, " ")

	return str
}

func (c *Client) formatMessage(chatID int64, messageID int64, preview bool, message *client.Message) string {
	var err error
	if message == nil {
		message, err = c.client.GetMessage(&client.GetMessageRequest{
			ChatId:    chatID,
			MessageId: messageID,
		})
		if err != nil {
			return "<error fetching message>"
		}
	}

	if message == nil {
		return ""
	}

	var str strings.Builder
	str.WriteString(fmt.Sprintf("%v | %s | ", message.Id, c.formatContact(int64(message.SenderUserId))))
	// TODO: timezone
	if !preview {
		str.WriteString(time.Unix(int64(message.Date), 0).UTC().Format("02 Jan 2006 15:04:05 | "))
	}

	var text string
	switch message.Content.MessageContentType() {
	case client.TypeMessageText:
		messageText, _ := message.Content.(*client.MessageText)
		text = messageText.Text.Text
		// TODO: handle other message types with labels (not supported in Zhabogram!)
	}
	if text != "" {
		if !preview {
			str.WriteString(text)
		} else {
			newlinePos := strings.Index(text, newlineChar)
			if !preview || newlinePos == -1 {
				str.WriteString(text)
			} else {
				str.WriteString(text[0:newlinePos])
			}
		}
	}

	return str.String()
}

func (c *Client) formatContent(file *client.File, filename string) string {
	if file == nil {
		return ""
	}

	return fmt.Sprintf(
		"%s (%v kbytes) | %s/%s%s",
		filename,
		file.Size/1024,
		c.content.Link,
		fmt.Sprintf("%x", sha256.Sum256([]byte(file.Remote.Id))),
		filepath.Ext(filename),
	)
}

func (c *Client) messageToText(message *client.Message) string {
	switch message.Content.MessageContentType() {
	case client.TypeMessageSticker:
		sticker, _ := message.Content.(*client.MessageSticker)
		return sticker.Sticker.Emoji
	case client.TypeMessageBasicGroupChatCreate, client.TypeMessageSupergroupChatCreate:
		return "has created chat"
	case client.TypeMessageChatJoinByLink:
		return "joined chat via invite link"
	case client.TypeMessageChatAddMembers:
		addMembers, _ := message.Content.(*client.MessageChatAddMembers)

		text := "invited "
		if len(addMembers.MemberUserIds) > 0 {
			text += c.formatContact(int64(addMembers.MemberUserIds[0]))
		}

		return text
	case client.TypeMessageChatDeleteMember:
		deleteMember, _ := message.Content.(*client.MessageChatDeleteMember)
		return "kicked " + c.formatContact(int64(deleteMember.UserId))
	case client.TypeMessagePinMessage:
		pinMessage, _ := message.Content.(*client.MessagePinMessage)
		return "pinned message: " + c.formatMessage(message.ChatId, pinMessage.MessageId, false, nil)
	case client.TypeMessageChatChangeTitle:
		changeTitle, _ := message.Content.(*client.MessageChatChangeTitle)
		return "chat title set to: " + changeTitle.Title
	case client.TypeMessageLocation:
		location, _ := message.Content.(*client.MessageLocation)
		return fmt.Sprintf(
			"coordinates: %v,%v | https://www.google.com/maps/search/%v,%v/",
			location.Location.Latitude,
			location.Location.Longitude,
			location.Location.Latitude,
			location.Location.Longitude,
		)
	case client.TypeMessagePhoto:
		photo, _ := message.Content.(*client.MessagePhoto)
		return photo.Caption.Text
	case client.TypeMessageAudio:
		audio, _ := message.Content.(*client.MessageAudio)
		return audio.Caption.Text
	case client.TypeMessageVideo:
		video, _ := message.Content.(*client.MessageVideo)
		return video.Caption.Text
	case client.TypeMessageDocument:
		document, _ := message.Content.(*client.MessageDocument)
		return document.Caption.Text
	case client.TypeMessageText:
		text, _ := message.Content.(*client.MessageText)
		return text.Text.Text
	case client.TypeMessageVoiceNote:
		voice, _ := message.Content.(*client.MessageVoiceNote)
		return voice.Caption.Text
	case client.TypeMessageVideoNote:
		return ""
	case client.TypeMessageAnimation:
		animation, _ := message.Content.(*client.MessageAnimation)
		return animation.Caption.Text
	}

	return fmt.Sprintf("unknown message (%s)", message.Content.MessageContentType())
}

func (c *Client) contentToFilename(content client.MessageContent) (*client.File, string) {
	switch content.MessageContentType() {
	case client.TypeMessageSticker:
		sticker, _ := content.(*client.MessageSticker)
		return sticker.Sticker.Sticker, "sticker.webp"
	case client.TypeMessageVoiceNote:
		voice, _ := content.(*client.MessageVoiceNote)
		return voice.VoiceNote.Voice, fmt.Sprintf("voice note (%v s.).oga", voice.VoiceNote.Duration)
	case client.TypeMessageVideoNote:
		video, _ := content.(*client.MessageVideoNote)
		return video.VideoNote.Video, fmt.Sprintf("video note (%v s.).mp4", video.VideoNote.Duration)
	case client.TypeMessageAnimation:
		animation, _ := content.(*client.MessageAnimation)
		return animation.Animation.Animation, "animation.mp4"
	case client.TypeMessagePhoto:
		photo, _ := content.(*client.MessagePhoto)
		sizes := photo.Photo.Sizes
		if len(sizes) > 1 {
			file := sizes[len(sizes)-1].Photo
			return file, strconv.FormatInt(int64(file.Id), 10) + ".jpg"
		}
		return nil, ""
	case client.TypeMessageAudio:
		audio, _ := content.(*client.MessageAudio)
		return audio.Audio.Audio, audio.Audio.FileName
	case client.TypeMessageVideo:
		video, _ := content.(*client.MessageVideo)
		return video.Video.Video, video.Video.FileName
	case client.TypeMessageDocument:
		document, _ := content.(*client.MessageDocument)
		return document.Document.Document, document.Document.FileName
	}

	return nil, ""
}

func (c *Client) messageToPrefix(message *client.Message, fileString string) string {
	prefix := []string{}
	// message direction
	var directionChar string
	if message.IsOutgoing {
		directionChar = "➡ "
	} else {
		directionChar = "⬅ "
	}
	prefix = append(prefix, directionChar+strconv.FormatInt(message.Id, 10))
	// show sender in group chats
	if message.ChatId < 0 && message.SenderUserId != 0 {
		prefix = append(prefix, c.formatContact(int64(message.SenderUserId)))
	}
	if message.ForwardInfo != nil {
		switch message.ForwardInfo.Origin.MessageForwardOriginType() {
		case client.TypeMessageForwardOriginUser:
			originUser := message.ForwardInfo.Origin.(*client.MessageForwardOriginUser)
			prefix = append(prefix, "fwd: "+c.formatContact(int64(originUser.SenderUserId)))
		case client.TypeMessageForwardOriginHiddenUser:
			originUser := message.ForwardInfo.Origin.(*client.MessageForwardOriginHiddenUser)
			prefix = append(prefix, fmt.Sprintf("fwd: anonymous (%s)", originUser.SenderName))
		case client.TypeMessageForwardOriginChannel:
			channel := message.ForwardInfo.Origin.(*client.MessageForwardOriginChannel)
			var signature string
			if channel.AuthorSignature != "" {
				signature = fmt.Sprintf(" (%s)", channel.AuthorSignature)
			}
			prefix = append(prefix, "fwd: "+c.formatContact(channel.ChatId)+signature)
		}
	}
	// reply to
	if message.ReplyToMessageId != 0 {
		prefix = append(prefix, "reply: "+c.formatMessage(message.ChatId, message.ReplyToMessageId, true, nil))
	}
	if fileString != "" {
		prefix = append(prefix, "file: "+fileString)
	}

	return strings.Join(prefix, " | ")
}

// ProcessOutgoingMessage executes commands or sends messages to mapped chats
func (c *Client) ProcessOutgoingMessage(chatID int64, text string, messageID int64, returnJid string) {
	if strings.HasPrefix(text, "/") {
		// try to execute a command
		response, isCommand := c.ProcessChatCommand(text)
		if response != "" {
			gateway.SendMessage(returnJid, strconv.FormatInt(chatID, 10), response, c.xmpp)
		}
		// do not send on success
		if isCommand {
			return
		}
	}

	if !c.online {
		// we're offline
		return
	}

	log.Warnf("Send message to chat %v", chatID)

	// quotations
	var reply int64
	replySlice := replyRegex.FindStringSubmatch(text)
	if len(replySlice) > 1 {
		reply, _ = strconv.ParseInt(replySlice[1], 10, 64)
	}

	// attach a file
	var file *client.InputFileRemote
	if c.content.Upload != "" && strings.HasPrefix(text, c.content.Upload) {
		file = &client.InputFileRemote{
			Id: text,
		}
	}

	// remove first line from text
	if file != nil || reply != 0 {
		newlinePos := strings.Index(text, newlineChar)
		if newlinePos != -1 {
			text = text[newlinePos+1:]
		}
	}
	formattedText := &client.FormattedText{
		Text: text,
	}

	var message client.InputMessageContent
	if file != nil {
		// we can try to send a document
		message = &client.InputMessageDocument{
			Document: file,
			Caption:  formattedText,
		}
	} else {
		// compile our message
		message = &client.InputMessageText{
			Text: formattedText,
		}
	}

	if messageID != 0 {
		c.client.EditMessageText(&client.EditMessageTextRequest{
			ChatId:              chatID,
			MessageId:           messageID,
			InputMessageContent: message,
		})
	} else {
		c.client.SendMessage(&client.SendMessageRequest{
			ChatId:              chatID,
			ReplyToMessageId:    reply,
			InputMessageContent: message,
		})
	}
}
